import createPub from '../';
import {usePub}  from '../';

type QueryFuncType = (...args: Array<any>) => Promise;

type CreateQueryPubOptions = {
    queryFunc?: QueryFuncType,
    cacheTimeout?: number,
    hidePendingOnRefresh?: boolean
    avoidRepeatInProgress?: boolean,
    forceRefreshAfterTimeout?: boolean,
}

enum PromiseStatus  {
    PENDING = "PENDING",
    FULFILLED = "FULFILLED",
    REJECTED = "REJECTED"
}

type PromiseSnapshot = {
    status: PromiseStatus,
    payload: any,
    error: Error
}

const defaultOptions:CreateQueryPubOptions = {
    cacheTimeout: 2 * 60 * 1000,
    hidePendingOnRefresh: true,
    avoidRepeatInProgress: true,
    forceRefreshAfterTimeout: true,
}

export default (options:CreateQueryPubOptions) => {
    const queryPub = createPub({queries: {}, mutations: {}})
    const currentPromises = {queries: {}, mutations: {}}

    options = {...defaultOptions, ...options}

    const withQueryFunc = (queryFunc?: QueryFuncType):Function => {
        return ({promiseName, isMutation, cacheTagFunc, args=[]}) => {
            const qF = queryFunc || options.queryFunc
            const defaultArgs = args

            const mixArgs = (args:any[]) => 
                Array.isArray(defaultArgs) ? 
                    [...defaultArgs, ...args] :
                    (typeof defaultArgs === 'function' ? 
                        defaultArgs(...args) : args)

            const stringifyArgs = args => 
                JSON.stringify(mixArgs(args))

            const makeQuery = (...args:any[]) => {
                return qF(...mixArgs(args))
            }

            const intersectTags = (first:Array<any>, second:Array<any>) => 
                first.find(firstItem => second.find(secondItem => firstItem.type === secondItem.type && firstItem.id === secondItem.id))


            const invalidateTags = tags => {
                console.log(tags)
                if (!tags) return
                for (const promiseBranch of Object.values(queryPub.queries)){
                    for (const [args, promiseState] of Object.entries(promiseBranch))
                        if (promiseState.cacheTags && intersectTags(tags, promiseState.cacheTags)){
                            console.log('invalidate', tags, args, promiseState)
                            delete promiseBranch[args]
                        }
                }
            }

            const query = (...args:any[]) => {
                if (isMutation){
                    if (options.avoidRepeatInProgress && queryPub.mutations[promiseName]?.refreshing){
                        return currentPromises.mutations[promiseName]
                    }
                    else {
                        if (!options.hidePendingOnRefresh || !queryPub.mutations[promiseName]){
                            queryPub.mutations[promiseName] = {status: 'PENDING', refreshing: true}
                        }
                        if (!queryPub.mutations[promiseName].refreshing){
                            queryPub.mutations[promiseName].refreshing = true
                        }
                        return currentPromises.mutations[promiseName] = makeQuery(...args).then(payload => queryPub.mutations[promiseName] = (invalidateTags(cacheTagFunc && cacheTagFunc(payload)),{status: 'FULFILLED', payload}),
                                                                                                error   => queryPub.mutations[promiseName] = {status: 'REJECTED', error})
                    }
                }
                if (!isMutation){
                    const stringifiedArgs = stringifyArgs(args)
                    if ((!options.cacheTimeout && queryPub.queries[promiseName]?.[stringifiedArgs].status === 'FULFILLED') || 
                        (queryPub.queries[promiseName]?.[stringifiedArgs] && (Date.now() - queryPub.queries[promiseName][stringifiedArgs].timeStamp) < options.cacheTimeout)){

                        return currentPromises.queries[promiseName][stringifiedArgs] 
                    }
                    return forceQuery(...args)
                }
            }

            const forceQuery = (...args:any[]) =>  {
                if (isMutation) return query(...args)



                const stringifiedArgs = stringifyArgs(args)

                if (!queryPub.queries[promiseName]) {
                    queryPub.queries[promiseName] = {}
                }
                if (!currentPromises.queries[promiseName]){
                    currentPromises.queries[promiseName] = {}
                }



                if (options.avoidRepeatInProgress && queryPub.queries[promiseName]?.[stringifiedArgs]?.refreshing){
                    return currentPromises.queries[promiseName][stringifiedArgs]
                }

                if (!options.hidePendingOnRefresh || !queryPub.queries[promiseName]?.[stringifiedArgs]){
                    queryPub.queries[promiseName][stringifiedArgs] = {status: 'PENDING', timeStamp: Date.now(), refreshing: true}
                }

                if (!queryPub.queries[promiseName][stringifiedArgs].refreshing){
                    queryPub.queries[promiseName][stringifiedArgs].refreshing = true
                }

                if (queryPub.queries[promiseName][stringifiedArgs].timeoutId){
                    clearTimeout(queryPub.queries[promiseName][stringifiedArgs].timeoutId)
                }



                return currentPromises.queries[promiseName][stringifiedArgs] = makeQuery(...args)
                                                                                    .then(payload => queryPub.queries[promiseName][stringifiedArgs] = {    
                                                                                                                                                            status: 'FULFILLED', 
                                                                                                                                                            payload, 
                                                                                                                                                            timeStamp: Date.now(), 
                                                                                                                                                            cacheTags: cacheTagFunc(payload),
                                                                                                                                                            timeoutId: options.cacheTimeout && options.forceRefreshAfterTimeout ? setTimeout(() =>forceQuery(...args), options.cacheTimeout) : null
                                                                                                                                                        },
                                                                                           error   => queryPub.queries[promiseName][stringifiedArgs] = {
                                                                                                                                                            status: 'REJECTED', 
                                                                                                                                                            error, 
                                                                                                                                                            timeStamp: Date.now(),
                                                                                                                                                            timeoutId: options.cacheTimeout && options.forceRefreshAfterTimeout ? setTimeout(() =>forceQuery(...args), options.cacheTimeout) : null
                                                                                                                                                        })
            }

            const hookName = `use${promiseName[0].toUpperCase() + promiseName.slice(1) + (isMutation ? 'Mutation' : 'Query')}`

            const useMutation = () => {
                usePub(queryPub.mutations)

                return [query, queryPub.mutations[promiseName] || {status: 'INIT'}]
            }

            const useQuery = (...args) => {
                query(...args)

                usePub(queryPub.queries[promiseName])

                const stringifiedArgs = stringifyArgs(args)
                return queryPub.queries[promiseName][stringifiedArgs]
            }

            return {makeQuery, query, forceQuery, [hookName]: isMutation ? useMutation : useQuery}
        }
    }

    return {queryPub, withQueryFunc, ...(options.queryFunc ? {createQuery: withQueryFunc(options.queryFunc)}: {})}
}
